#!/usr/bin/env python
"""
#   GAL/PAL chip emulator GUI
#
#   Copyright (C) 2008 Michael Buesch <mb@bu3sch.de>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import galemu
import sys
import ConfigParser
from PyQt4.QtCore import *
from PyQt4.QtGui import *


class AbstractInput(QObject):
	def __init__(self, parent, abstractDevice):
		QObject.__init__(self, parent)
		self.abstractDevice = abstractDevice

	def changeState(self, newState):
		"""Set the state of this input.
		   Reimplement me in the subclass!"""
		raise Exception("AbstractInput.changeState() unimplemented")

	def getDevice(self):
		return self.abstractDevice

	def connectToOutput(self, output):
		self.connect(output, SIGNAL("stateChanged(QObject, int)"),
			     self.__connectedOutputChangedState)

	def disconnectFromOutput(self, output):
		self.disconnect(output, SIGNAL("stateChanged(QObject, int)"),
				self.__connectedOutputChangedState)

	def __connectedOutputChangedState(self, output, newState):
		self.changeState(newState)

class AbstractOutput(QObject):
	def __init__(self, parent, abstractDevice):
		"Emits: stateChanged(self, newState)"
		QObject.__init__(self, parent)
		self.abstractDevice = abstractDevice

	def getDevice(self):
		return self.abstractDevice

	def connectToInput(self, input):
		input.connectToOutput(self)

	def disconnectFromInput(self, input):
		input.disconnectFromOutput(self)

	def changeNotify(self, newState):
		self.emit(SIGNAL("stateChanged(QObject, int)"), self, newState)

class AbstractDevice(QObject):
	def __init__(self):
		QObject.__init__(self)

class Oscilloscope(QWidget, AbstractDevice):
	def __init__(self, parent, flags = Qt.Window):
		AbstractDevice.__init__(self)
		QWidget.__init__(self, parent, flags)
		self.mainwidget = parent
		self.setWindowTitle("Oscilloscope")
		#TODO
		QLabel("Not implemented", self)

	def closeEvent(self, ev):
		self.mainwidget.removeScope(self)
		self.mainwidget.updateAllOutputReferences()

class GeneratorOutput(QWidget, AbstractOutput):
	def __init__(self, signalGenerator, pinNr):
		AbstractOutput.__init__(self, signalGenerator, signalGenerator)
		QWidget.__init__(self, signalGenerator)
		self.pinNr = pinNr
		self.connectedInput = None

		layout = QHBoxLayout()
		self.cb = QCheckBox("Out %d" % pinNr, self)
		self.cb.setEnabled(False)
		layout.addWidget(self.cb)
		self.combo = QComboBox(self)
		self.connect(self.combo, SIGNAL("currentIndexChanged(int)"),
			     self.outputRefChanged)
		layout.addWidget(self.combo)

		self.setLayout(layout)

	def setState(self, state):
		oldState = (self.cb.checkState() != Qt.Unchecked)
		if oldState == state:
			return
		if state:
			self.cb.setCheckState(Qt.Checked)
		else:
			self.cb.setCheckState(Qt.Unchecked)
		self.changeNotify(state)

	def getState(self):
		return self.cb.checkState() != Qt.Unchecked

	def setReferences(self, refs):
		mainwidget = self.getDevice().mainwidget
		selectedInput = None
		if self.combo.currentIndex() > 0:
			# Get the "class OneInput" of the old selection
			variant = self.combo.itemData(self.combo.currentIndex())
			selectedInput = mainwidget.idstrToInput(variant.toString())
		self.combo.clear()
		for ref in refs:
			self.combo.addItem(ref[0], QVariant(ref[1]))
			self.combo.adjustSize()
		if selectedInput:
			idstr = mainwidget.inputToIdstr(selectedInput)
			idx = self.combo.findData(QVariant(idstr))
			if idx >= 0:
				self.combo.setCurrentIndex(idx)

	def outputRefChanged(self, newIndex):
		if newIndex < 0:
			return
		if newIndex > 0:
			signalGenerator = self.getDevice()
			mainwidget = signalGenerator.mainwidget
			newInput = mainwidget.idstrToInput(self.combo.itemData(newIndex).toString())
			self.__disconnectCurrentRef()
			self.connectToInput(newInput)
			self.connectedInput = newInput
			self.connectedInput.setState(self.getState())
		else:
			self.__disconnectCurrentRef()

	def __disconnectCurrentRef(self):
		if not self.connectedInput:
			return
		curInput = self.connectedInput
		self.disconnectFromInput(curInput)
		self.connectedInput = None

class GeneratorInterpreter:
	class KillException:
		pass

	def __init__(self, signalGenerator, script):
		self.signalGenerator = signalGenerator
		self.script = str(script)
		self.killRequest = False

	def callback_relax(self):
		"Just relax a bit and let the user interface refresh"
		if self.killRequest:
			raise GeneratorInterpreter.KillException
		app.processEvents()

	def callback_msleep(self, msec):
		"Suspend the generator script for a few microseconds"
		start = QDateTime.currentDateTime()
		end = start.addMSecs(msec)
		while end > QDateTime.currentDateTime():
			self.callback_relax()
			QThread.msleep(1)

	def callback_out(self, pin, newState):
		"Set the state of an output pin"
		self.callback_relax()
		outputs = self.signalGenerator.getOutputs()
		try:
			generatorOutput = outputs[pin]
		except IndexError:
			QMessageBox.critical(None, "OUT invalid parameter",
					     "pinNumber parameter to out(pinNumber, newState) out of range.")
			raise GeneratorInterpreter.KillException
		except TypeError:
			QMessageBox.critical(None, "OUT invalid parameter",
					     "Invalid parameter to out(pinNumber, newState) function.")
			raise GeneratorInterpreter.KillException
		generatorOutput.setState(newState)

	def run(self):
		context = {
			"relax" : self.callback_relax,
			"msleep" : self.callback_msleep,
			"out" : self.callback_out,
		}
		try:
			exec self.script in context
		except GeneratorInterpreter.KillException:
			pass
		except Exception, e:
			QMessageBox.critical(None, "Generator script failed",
					     "The Generator script failed with an exception:\n" +\
					     e.message)

	def kill(self):
		self.killRequest = True

class GeneratorScriptEdit(QTextEdit):
	def __init__(self, parent):
		QTextEdit.__init__(self, parent)
		self.setFontFamily("Courier")
		self.setFontPointSize(12)

	def text(self):
		return self.toPlainText()

class SignalGenerator(QWidget, AbstractDevice):
	def __init__(self, parent, nrOutputs, flags = Qt.Window):
		AbstractDevice.__init__(self)
		QWidget.__init__(self, parent, flags)
		self.mainwidget = parent
		self.running = False
		self.nrOutputs = nrOutputs
		self.setWindowTitle("Signal generator")

		layout = QHBoxLayout()

		v = QVBoxLayout()
		h = QHBoxLayout()
		self.loadButton = QPushButton("Load file...", self)
		self.connect(self.loadButton, SIGNAL("clicked()"),
			     self.loadFile)
		h.addWidget(self.loadButton)
		self.saveButton = QPushButton("Save to file...", self)
		self.connect(self.saveButton, SIGNAL("clicked()"),
			     self.saveToFile)
		h.addWidget(self.saveButton)
		v.addLayout(h)

		self.script = GeneratorScriptEdit(self)
		v.addWidget(self.script)

		h = QHBoxLayout()
		self.runButton = QPushButton("Run...", self)
		self.connect(self.runButton, SIGNAL("clicked()"),
			     self.runScript)
		h.addWidget(self.runButton)
		self.killButton = QPushButton("Kill", self)
		self.connect(self.killButton, SIGNAL("clicked()"),
			     self.killScript)
		self.killButton.setEnabled(False)
		h.addWidget(self.killButton)
		self.helpButton = QPushButton("Help", self)
		self.connect(self.helpButton, SIGNAL("clicked()"),
			     self.showHelp)
		h.addWidget(self.helpButton)
		v.addLayout(h)
		layout.addLayout(v)

		v = QVBoxLayout()
		self.outputs = []
		for i in range(0, nrOutputs):
			outp = GeneratorOutput(self, i)
			v.addWidget(outp)
			self.outputs.append(outp)
		layout.addLayout(v)

		self.setLayout(layout)

		self.updateOutputReferences()

	def getOutputs(self):
		return self.outputs

	def closeEvent(self, ev):
		self.killScript()
		self.mainwidget.removeGenerator(self)
		self.mainwidget.updateAllOutputReferences()

	def loadFile(self):
		fn = QFileDialog.getOpenFileName(self, "Open Galemu generator script")
		if not fn:
			return
		try:
			self.script.setText(file(fn).read())
		except IOError, e:
			QMessageBox.critical(self, "Failed to open generator script",
					     "Failed to open generator script %s: %s" %\
					     (fn, e.strerror))

	def saveToFile(self):
		fn = QFileDialog.getSaveFileName(self, "Save Galemu generator script")
		if not fn:
			return
		try:
			fd = file(fn, "w")
			fd.write(self.script.text())
		except IOError, e:
			QMessageBox.critical(self, "Failed to save generator script",
					     "Failed to save generator script %s: %s" %\
					     (fn, e.strerror))

	def updateGui(self):
		self.runButton.setEnabled(not self.running)
		self.killButton.setEnabled(self.running)
		self.script.setEnabled(not self.running)
		self.loadButton.setEnabled(not self.running)

	def runScript(self):
		self.interp = GeneratorInterpreter(self, self.script.text())
		self.running = True
		self.updateGui()
		self.interp.run()
		self.running = False
		self.updateGui()

	def killScript(self):
		if not self.running:
			return
		self.interp.kill()

	def showHelp(self):
		t = "The Galemu generator script is an ordinary python script with a few "+\
		"additional special functions:\n\n"+\
		"relax()  : Call this to give the Galemu GUI a chance to refresh.\n"+\
		"msleep(msec)  : Sleep for \"msec\" milliseconds.\n"+\
		"out(pin, newState)  : Set the output \"pin\" to a new boolean \"newState\"."
		QMessageBox.information(self, "Galemu generator script quickhelp", t)

	def updateOutputReferences(self):
		refs = [ ("=>     No connection     ", ":"), ]
		devices = self.mainwidget.getDeviceList()
		for dev in devices:
			inputsWidget = dev.getInputsWidget()
			inputsList = inputsWidget.getInputsList()
			if not inputsList:
				continue # Dev is not initialized
			for input in inputsList:
				namestr = "=> %s - %s" % (dev.getNameString(), input.getNameString())
				idstr = self.mainwidget.inputToIdstr(input)
				refs.append( (namestr, idstr) )
		for output in self.getOutputs():
			output.setReferences(refs)

class OneInput(QWidget, AbstractInput):
	def __init__(self, parent, inpin):
		AbstractInput.__init__(self, parent, parent.deviceWidget)
		QWidget.__init__(self, parent)
		self.inputsWidget = parent
		self.inpin = inpin

		layout = QHBoxLayout()
		self.cb = QCheckBox("In %d" % inpin, self)
		self.connect(self.cb, SIGNAL("stateChanged(int)"),
			     self.stateChanged)
		layout.addWidget(self.cb)

		self.setLayout(layout)

	def getPinNumber(self):
		return self.inpin

	def getNameString(self):
		return self.cb.text()

	def stateChanged(self, newState):
		self.emit(SIGNAL("stateChanged(int)"), newState)

	def getState(self):
		if self.cb.checkState() == Qt.Unchecked:
			return False
		return True

	def setState(self, newState):
		if newState:
			if self.cb.checkState() == Qt.Unchecked:
				self.cb.toggle()
		else:
			if self.cb.checkState() != Qt.Unchecked:
				self.cb.toggle()

	def changeState(self, newState):
		"Reimplements AbstractInput.changeState()"
		self.setState(newState)
		if not self.inputsWidget.deviceWidget.isFrozen():
			self.inputsWidget.commit()

class InputsWidget(QWidget):
	def __init__(self, parent):
		QWidget.__init__(self, parent)
		self.deviceWidget = parent

		self.setLayout(QVBoxLayout())
		self.inputs = []
		self.emu = None

	def getInputsList(self):
		return self.inputs

	def setup(self, emu):
		self.emu = emu
		if self.inputs:
			self.destroy()
		gal = emu.getGal()
		nrInput = len(gal.getInputList())
		self.inputs = []
		for i in range(0, nrInput):
			inpin = gal.getInputList()[i].getPin()
			inp = OneInput(self, inpin)
			self.connect(inp, SIGNAL("stateChanged(int)"),
				     self.oneStateChanged)
			self.inputs.append(inp)
			self.layout().addWidget(inp)

	def destroy(self):
		for i in self.inputs:
			i.deleteLater()
		self.inputs = []

	def oneStateChanged(self, state):
		devicewidget = self.parent()
		if devicewidget.isFrozen():
			return
		self.commit()
		devicewidget.getOutputsWidget().update()

	def commit(self):
		inpStates = []
		# We depend on self.inputs being sorted like emu expects it.
		for inp in self.inputs:
			inpStates.append(inp.getState())
		try:
			self.emu.setInput(inpStates)
		except galemu.GalemuEx, e:
			QMessageBox.warning(self, "Emulator error",
					    e.message)
			return

	def reset(self):
		devicewidget = self.parent()
		for inp in self.inputs:
			inp.setState(False)
		if not devicewidget.isFrozen():
			self.commit()

	def getInputByPinNumber(self, pinnr):
		for inp in self.inputs:
			if inp.getPinNumber() == pinnr:
				return inp
		return None

	def save(self):
		text = ""
		for inp in self.inputs:
			stat = 0
			if inp.getState():
				stat = 1
			text += "input%d=%d\n" % (inp.getPinNumber(), stat)
		return text

	def setInputPinState(self, pinNr, newState):
		for inp in self.inputs:
			if inp.getPinNumber() == pinNr:
				inp.setState(newState)
				break

class OneOutput(QWidget, AbstractOutput):
	def __init__(self, parent, outpin):
		AbstractOutput.__init__(self, parent, parent.deviceWidget)
		QWidget.__init__(self, parent)
		self.outputsWidget = parent
		self.outpin = outpin
		self.connectedInput = None

		layout = QHBoxLayout()
		self.cb = QCheckBox("Out %d" % outpin, self)
		self.cb.setEnabled(False)
		layout.addWidget(self.cb)
		self.combo = QComboBox(self)
		self.connect(self.combo, SIGNAL("currentIndexChanged(int)"),
			     self.outputRefChanged)
		layout.addWidget(self.combo)

		self.setLayout(layout)

	def getPinNumber(self):
		return self.outpin

	def getConnectedInput(self):
		return self.connectedInput

	def setState(self, state):
		oldState = (self.cb.checkState() != Qt.Unchecked)
		if oldState == state:
			return
		if state:
			self.cb.setCheckState(Qt.Checked)
		else:
			self.cb.setCheckState(Qt.Unchecked)
		self.changeNotify(state)

	def getState(self):
		return self.cb.checkState() != Qt.Unchecked

	def setReferences(self, refs):
		mainwidget = self.outputsWidget.deviceWidget.mainwidget
		selectedInput = None
		if self.combo.currentIndex() > 0:
			# Get the "class OneInput" of the old selection
			variant = self.combo.itemData(self.combo.currentIndex())
			selectedInput = mainwidget.idstrToInput(variant.toString())
		self.combo.clear()
		for ref in refs:
			self.combo.addItem(ref[0], QVariant(ref[1]))
			self.combo.adjustSize()
		if selectedInput:
			idstr = mainwidget.inputToIdstr(selectedInput)
			idx = self.combo.findData(QVariant(idstr))
			if idx >= 0:
				self.combo.setCurrentIndex(idx)

	def outputRefChanged(self, newIndex):
		if newIndex < 0:
			return
		if newIndex > 0:
			deviceWidget = self.outputsWidget.deviceWidget
			mainwidget = deviceWidget.mainwidget
			newInput = mainwidget.idstrToInput(self.combo.itemData(newIndex).toString())
			self.__disconnectCurrentRef()
			self.connectToInput(newInput)
			self.connectedInput = newInput
			self.connectedInput.setState(self.getState())
		else:
			self.__disconnectCurrentRef()

	def __disconnectCurrentRef(self):
		if not self.connectedInput:
			return
		curInput = self.connectedInput
		self.disconnectFromInput(curInput)
		self.connectedInput = None

	def forceConnectReference(self, select_idstr):
		for i in range(0, self.combo.count()):
			idstr = str(self.combo.itemData(i).toString())
			if idstr == select_idstr:
				self.combo.setCurrentIndex(i)
				break

class OutputsWidget(QWidget):
	def __init__(self, parent):
		QWidget.__init__(self, parent)

		self.setLayout(QVBoxLayout())
		self.outputs = []
		self.emu = None
		self.deviceWidget = parent

	def getOutputsList(self):
		return self.outputs

	def setup(self, emu):
		self.emu = emu
		if self.outputs:
			self.destroy()
		gal = emu.getGal()
		nrOutput = len(gal.getOlmcList())
		self.outputs = []
		for i in range(0, nrOutput):
			outpin = gal.getOlmcList()[i].getOutputPin().getPin()
			o = OneOutput(self, outpin)
			self.outputs.append(o)
			self.layout().addWidget(o)
		self.update()

	def destroy(self):
		for o in self.outputs:
			o.deleteLater()
		self.outputs = []

	def update(self):
		try:
			out = self.emu.getOutput()
		except galemu.GalemuEx, e:
			QMessageBox.warning(self, "Emulator error",
					    e.message)
			return
		# We depend on emu.getOutput() being sorted like we expect it.
		for i in range(0, len(out)):
			self.outputs[i].setState(out[i])

	def updateReferences(self):
		refs = [ ("=>     No connection     ", ":"), ]
		devices = self.deviceWidget.mainwidget.getDeviceList()
		for dev in devices:
			if dev is self.deviceWidget:
				continue # Skip myself
			inputsWidget = dev.getInputsWidget()
			inputsList = inputsWidget.getInputsList()
			if not inputsList:
				continue # Dev is not initialized
			for input in inputsList:
				namestr = "=> %s - %s" % (dev.getNameString(), input.getNameString())
				idstr = self.deviceWidget.mainwidget.inputToIdstr(input)
				refs.append( (namestr, idstr) )
		for output in self.outputs:
			output.setReferences(refs)

	def selectReference(self, pinNr, connTo_idstr):
		for output in self.outputs:
			if output.getPinNumber() == pinNr:
				output.forceConnectReference(connTo_idstr)

	def save(self):
		text = ""
		for output in self.outputs:
			connectedInput = output.getConnectedInput()
			if not connectedInput:
				continue
			idstr = self.deviceWidget.mainwidget.inputToIdstr(connectedInput)
			text += "outputconnection%d=\"%s\"\n" % (output.getPinNumber(), idstr)
		return text

class DeviceWidget(QWidget, AbstractDevice):
	def __init__(self, enum, parent, flags = Qt.Window):
		AbstractDevice.__init__(self)
		QWidget.__init__(self, parent, flags)
		self.mainwidget = parent

		layout = QVBoxLayout()

		h = QHBoxLayout()
		self.nameEdit = QLineEdit("Device %d" % enum, self)
		self.nameEdit.setEnabled(False)
		self.setWindowTitle(self.nameEdit.text())
		h.addWidget(self.nameEdit)
		self.nameEditEnable = QCheckBox("edit", self)
		self.connect(self.nameEditEnable, SIGNAL("toggled(bool)"),
			     self.nameEditEnableChanged)
		h.addWidget(self.nameEditEnable)
		layout.addLayout(h)

		devLayout = QHBoxLayout()
		devLayout.addStretch()

		self.inputsWidget = InputsWidget(self)
		devLayout.addWidget(self.inputsWidget)

		v = QVBoxLayout()
		v.addStretch()
		self.jedButton = QPushButton("\n\n\nL\no\na\nd\n \nJ\nE\nD\nE\nC\n\n\n", self)
		self.connect(self.jedButton, SIGNAL("clicked()"),
			     self.loadJedec)
		v.addWidget(self.jedButton)
		v.addStretch()

		devLayout.addLayout(v)
		self.outputsWidget = OutputsWidget(self)
		devLayout.addWidget(self.outputsWidget)

		devLayout.addStretch()
		layout.addLayout(devLayout)

		self.setLayout(layout)

		self.running = False

	def isRunning(self):
		return self.running

	def closeEvent(self, ev):
		self.mainwidget.removeDevice(self)
		self.mainwidget.updateAllOutputReferences()

	def getNameString(self):
		return self.nameEdit.text()

	def forceSetNameString(self, newName):
		self.nameEdit.setText(newName)
		self.setWindowTitle(self.nameEdit.text())
		self.mainwidget.updateAllOutputReferences()

	def getOutputsWidget(self):
		return self.outputsWidget

	def getInputsWidget(self):
		return self.inputsWidget

	def getEmulator(self):
		if self.running:
			return self.emu
		return None

	def nameEditEnableChanged(self):
		if self.nameEditEnable.isChecked():
			self.nameEdit.setEnabled(True)
		else:
			self.nameEdit.setEnabled(False)
			self.setWindowTitle(self.nameEdit.text())
			self.mainwidget.updateAllOutputReferences()

	def loadJedec(self):
		fn = QFileDialog.getOpenFileName(self, "Open JEDEC file")
		if not fn:
			return
		try:
			jedecData = file(fn, "r").read()
		except IOError, e:
			QMessageBox.critical(self, "JEDEC load failed",
					     "Failed to load JEDEC file: %s" % e.strerror)
			return
		self.loadJedecData(jedecData)

	def loadJedecData(self, data):
		#FIXME error handling
		self.jedecData = data
		self.emulatorSetup()
		self.guiSetup()
		self.running = True

	def emulatorSetup(self):
		try:
			array = galemu.FuseArray()
			array.readJedecData(self.jedecData)
			gal = galemu.selectGalClassObject(array.getGalType()) (array)
			self.emu = galemu.Emulator(gal)
		except galemu.GalemuEx, e:
			QMessageBox.critical(self, "Emulator startup failed",
					     "Failed to initialize the emulator: %s" % e.message)

	def guiSetup(self):
		self.inputsWidget.setup(self.emu)
		self.outputsWidget.setup(self.emu)
		self.mainwidget.updateAllOutputReferences()

	def isFrozen(self):
		return self.mainwidget.isFrozen()

	def reset(self):
		if not self.running:
			return
		self.inputsWidget.reset()
		self.emu.reset()
		if not self.mainwidget.isFrozen():
			self.outputsWidget.update()

	def updateState(self):
		if not self.running:
			return
		self.inputsWidget.commit()
		self.outputsWidget.update()

	def save(self):
		text = "jedec=" + QByteArray(self.jedecData).toHex() + "\n"
		text += self.inputsWidget.save()
		text += self.outputsWidget.save()
		return text

class MainWidget(QWidget):
	def __init__(self, parent):
		QWidget.__init__(self, parent)

		self.statusBar = parent.statusBar()
		self.devices = []
		self.devicesEnum = 0
		self.resetInProgress = False
		self.scopes = []
		self.generators = []

		layout = QVBoxLayout()

		h = QHBoxLayout()
		self.freezeButton = QPushButton("Global freeze", self)
		self.freezeButton.setCheckable(True)
		self.connect(self.freezeButton, SIGNAL("clicked()"),
			     self.freezeButtonClicked)
		h.addWidget(self.freezeButton)
		self.resetButton = QPushButton("Global reset", self)
		self.connect(self.resetButton, SIGNAL("clicked()"),
			     self.resetClicked)
		h.addWidget(self.resetButton)
		self.addButton = QPushButton("Add device", self)
		self.connect(self.addButton, SIGNAL("clicked()"),
			     self.addDevice)
		h.addWidget(self.addButton)
		layout.addLayout(h)
		h = QHBoxLayout()
		self.scopeButton = QPushButton("Add oscilloscope", self)
		self.connect(self.scopeButton, SIGNAL("clicked()"),
			     self.addScope)
		h.addWidget(self.scopeButton)
		self.generatorButton = QPushButton("Add signal generator", self)
		self.connect(self.generatorButton, SIGNAL("clicked()"),
			     self.addGenerator)
		h.addWidget(self.generatorButton)
		layout.addLayout(h)
		layout.addStretch()

		self.setLayout(layout)

	def isFrozen(self):
		if self.resetInProgress:
			return True
		return self.freezeButton.isChecked()

	def freezeButtonClicked(self):
		if not self.freezeButton.isChecked():
			for dev in self.devices:
				dev.updateState()

	def resetClicked(self):
		self.resetInProgress = True
		if self.freezeButton.isChecked():
			self.freezeButton.toggle()
		for dev in self.devices:
			dev.reset()
		self.resetInProgress = False
		for dev in self.devices:
			dev.updateState()

	def addScope(self):
		scope = Oscilloscope(self)
		self.scopes.append(scope)
		self.layout().addWidget(scope)
		return scope

	def removeScope(self, scope):
		self.scopes.remove(scope)
		scope.deleteLater()

	def addGenerator(self, nrOutputs = -1):
		if nrOutputs <= 0:
			(nrOutputs, ok) = QInputDialog.getInteger(self, "Number of outputs",
				"How many outputs should the new generator have?",
				12, 0, 100, 1)
			if not ok:
				return None
		generator = SignalGenerator(self, nrOutputs)
		self.generators.append(generator)
		self.layout().addWidget(generator)
		return generator

	def removeGenerator(self, generator):
		self.generators.remove(generator)
		generator.deleteLater()

	def addDevice(self):
		dev = DeviceWidget(self.devicesEnum, self)
		self.devicesEnum += 1
		self.devices.append(dev)
		self.layout().addWidget(dev)
		return dev

	def removeDevice(self, deviceWidget):
		self.devices.remove(deviceWidget)
		deviceWidget.deleteLater()

	def getDeviceList(self):
		return self.devices

	def updateAllOutputReferences(self):
		for dev in self.devices:
			dev.getOutputsWidget().updateReferences()
		for generator in self.generators:
			generator.updateOutputReferences()

	def idstrToInput(self, idstr):
		"""This is a hack. As QVariant() currently does not support custom types,
		   we use a string to identify a "class OneInput" globally via QVariant().
		   The string looks like "Devicename:pinnumber". """
		idstr = str(idstr) # Make sure it's a python string
		devName = idstr[: idstr.rfind(":")]
		pinNr = int( idstr[idstr.rfind(":")+1 :] )
		for dev in self.devices:
			if dev.getNameString() == devName:
				inp = dev.getInputsWidget().getInputByPinNumber(pinNr)
				if inp:
					return inp
		return None

	def inputToIdstr(self, oneInput):
		"""This is a hack. See idstrToInput()."""
		deviceWidget = oneInput.inputsWidget.deviceWidget
		idstr = "%s:%d" % (deviceWidget.getNameString(), oneInput.getPinNumber())
		return idstr

	def save(self):
		fn = QFileDialog.getSaveFileName(self, "Save emulator setup", QString(),
						 "GAL emulator state file (*.galemu)")
		if not fn:
			return
		fn = str(fn)
		if not fn.endswith(".galemu"):
			fn = fn + ".galemu"

		text = "# GAL emulator state file\n"
		text += QDateTime.currentDateTime().toString("'# Date: 'ddd MMMM d yyyy  hh:mm:ss\n")

		haveData = False
		for dev in self.devices:
			if not dev.isRunning():
				continue
			haveData = True
			text += "\n"
			text += "[DEV_%s]\n" % dev.getNameString()
			text += dev.save()
		if not haveData:
			QMessageBox.information(self, "Nothing to save",
						"There is nothing to save. First create and load a device.")
			return

		try:
			fd = file(fn, "w")
			fd.write(text)
		except IOError, e:
			QMessageBox.critical(self, "Failed to write file",
					     "Failed to write file %s: %s" % (fn, e.strerror))
			return
		self.statusBar.showMessage("Stored status to file %s" % fn)

	def load(self):
		haveRunningDev = False
		for dev in self.devices:
			if dev.isRunning():
				haveRunningDev = True
		if haveRunningDev:
			pressed = QMessageBox.question(self, "Discard current setup?",
						       "Loading a new setup will terminate all currently "+\
						       "running emulators. Do you want to continue?",
						       QMessageBox.Yes | QMessageBox.No)
			if pressed != QMessageBox.Yes:
				return
		fn = QFileDialog.getOpenFileName(self, "Load emulator setup", QString(),
						 "GAL emulator state file (*.galemu)\n" +\
						 "All files (*.*)")
		if not fn:
			return
		try:
			p = ConfigParser.ConfigParser()
			if not p.read([fn, ]):
				raise ConfigParser.Error()
		except ConfigParser.Error, e:
			QMessageBox.critical(self, "Failed to read file",
					     "Failed to read the state file %s" % fn)
			return

		# First kill all existing devices
		for dev in self.devices:
			self.removeDevice(dev)

		# Create the devices
		for section in p.sections():
			if not section.startswith("DEV_"):
				continue
			if not p.has_option(section, "jedec"):
				QMessageBox.critical(self, "No JEDEC data",
						     "No JEDEC data for %s available!" % section)
				continue
			jedecHex = p.get(section, "jedec")
			jedec = QByteArray.fromHex(jedecHex).data()
			dev = self.addDevice()
			dev.loadJedecData(jedec)
			dev.forceSetNameString(section[4:])
			# Set the input pin states
			for option in p.options(section):
				if not option.startswith("input"):
					continue
				try:
					pinNr = int(option[5:])
					state = p.getboolean(section, option)
				except ValueError:
					QMessageBox.critical(self, "Invalid input",
							     "Invalid input pin definition: %s" % option)
					continue
				dev.getInputsWidget().setInputPinState(pinNr, state)
		# Setup the output connections
		for dev in self.devices:
			section = str("DEV_" + dev.getNameString())
			for option in p.options(section):
				if not option.startswith("outputconnection"):
					continue
				try:
					pinNr = int(option[16:])
					connTo = p.get(section, option).strip("\"")
				except ValueError:
					QMessageBox.critical(self, "Invalid connection",
							     "Invalid output pin connection definition: %s"\
							     % option)
					continue
				dev.getOutputsWidget().selectReference(pinNr, connTo)
		self.statusBar.showMessage("Loaded status from file %s" % fn)

class StatusBar(QStatusBar):
	def showMessage(self, msg):
		QStatusBar.showMessage(self, msg, 3000)

class MainWindow(QMainWindow):
	def __init__(self, parent=None):
		QMainWindow.__init__(self, parent)
		self.setWindowTitle("GAL/PAL emulator version %s" % galemu.VERSION_STRING)

		mb = QMenuBar(self)
		emen = QMenu("Emulator", mb)
		emen.addAction("Load from file...", self.load)
		emen.addAction("Save to file...", self.save)
		emen.addSeparator()
		emen.addAction("Exit", self.close)
		mb.addMenu(emen)
		helpmen = QMenu("Help", mb)
		helpmen.addAction("About", self.about)
		mb.addMenu(helpmen)
		self.setMenuBar(mb)

		self.setStatusBar(StatusBar())
		self.setCentralWidget(MainWidget(self))

	def save(self):
		self.centralWidget().save()

	def load(self):
		self.centralWidget().load()

	def about(self):
		QMessageBox.information(self, "About",
					("GAL/PAL emulator version %s\n"
					 "Copyright (c) 2008 Michael Buesch") %\
					 galemu.VERSION_STRING)

	def closeEvent(self, ev):
		global app
		app.quit()

def main(argv):
	global app
	global mainwnd

	app = QApplication(argv)
	mainwnd = MainWindow()
	mainwnd.show()
	mainwnd.centralWidget().addDevice()

	return app.exec_()

if __name__ == "__main__":
	sys.exit(main(sys.argv))
